/******************************************************************************
	LED_SLOT.c
	For AVR AT tiny 2313
******************************************************************************/
#define	F_CPU	8000000UL

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<util/delay.h>

/******************************************************************************
	7SEG LED
******************************************************************************/
#define BLANK	0x00
#define SEG_A	0x40
#define SEG_B	0x20
#define SEG_C	0x10
#define SEG_D	0x08
#define SEG_E	0x04
#define SEG_F	0x02
#define SEG_G	0x01

char DRUM[]={
	// 0
	~(SEG_A),
	~(SEG_B | SEG_F | SEG_G),
	~(SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F),
	~(SEG_C | SEG_E | SEG_G ),
	~(SEG_D),
	// 1
	~(BLANK),
	~(SEG_B),
	~(SEG_B | SEG_C),
	~(SEG_C),
	~(BLANK),
	// 2
	~(SEG_A),
	~(SEG_A | SEG_F | SEG_G),
	~(SEG_A | SEG_B | SEG_D | SEG_E | SEG_G),
	~(SEG_C | SEG_D | SEG_G),
	~(SEG_D),
	// 3
	~(SEG_A),
	~(SEG_A | SEG_B | SEG_G),
	~(SEG_A | SEG_B | SEG_C | SEG_D | SEG_G),
	~(SEG_C | SEG_D | SEG_G),
	~(SEG_D),
	// 4
	~(BLANK),
	~(SEG_A | SEG_B),
	~(SEG_B | SEG_C | SEG_F | SEG_G),
	~(SEG_C | SEG_D | SEG_E),
	~(BLANK),
	// 5
	~(SEG_A),
	~(SEG_A | SEG_B | SEG_G),
	~(SEG_A | SEG_C | SEG_D | SEG_F | SEG_G),
	~(SEG_D | SEG_E | SEG_G),
	~(SEG_D),
	// 6
	~(SEG_A),
	~(SEG_A | SEG_B | SEG_F | SEG_G),
	~(SEG_A | SEG_C | SEG_D | SEG_E | SEG_F | SEG_G),
	~(SEG_D | SEG_E | SEG_G),
	~(SEG_D),
	// 7
	~(BLANK),
	~(SEG_B),
	~(SEG_A | SEG_B | SEG_C),
	~(SEG_C | SEG_G),
	~(SEG_D),
	// 8
	~(SEG_A),
	~(SEG_A | SEG_B | SEG_F | SEG_G),
	~(SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F | SEG_G),
	~(SEG_C | SEG_D | SEG_E | SEG_G),
	~(SEG_D),
	// 9
	~(SEG_A),
	~(SEG_A | SEG_B | SEG_G),
	~(SEG_A | SEG_B | SEG_C | SEG_D | SEG_F | SEG_G),
	~(SEG_C | SEG_D | SEG_E | SEG_G),
	~(SEG_D)
};

/******************************************************************************
	7SEG Display Function
******************************************************************************/
uint8_t g_DispBuff[3];

void disp7Seg(void)
{
	static uint8_t digit = 0;

	PORTD = (PORTD & 0x8F) | 0x70;
	PORTB = g_DispBuff[digit];
	switch(digit){
		case 0: PORTD = (PORTD & 0x8F) | 0x60; break;
		case 1: PORTD = (PORTD & 0x8F) | 0x50; break;
		case 2: PORTD = (PORTD & 0x8F) | 0x30; break;
	}
	digit++;
	if(digit > 2) digit = 0;
}

/******************************************************************************
	Timer Function
******************************************************************************/
#define COUNT_INTERVAL	20

volatile uint8_t timerCounter1ms = 0;
volatile uint8_t timerCounter20ms = 0;

ISR(TIMER0_COMPA_vect)
{
	disp7Seg();
	if(timerCounter1ms < COUNT_INTERVAL-1){
		timerCounter1ms++;
	}
	else{
		timerCounter1ms = 0;
		timerCounter20ms++;
	}
}

void initTimer(void)
{
	cli();	// Disable Interrupt

	// TCCR0A ... COM0A1 COM0A0 COM0B1 COM0B0 -      -      WGM01  WGM00
	// TCCR0B ... FC0A   FOC0B  -      -      WGM02  CS02   CS01   CS00
	//
	// WGM0[2:0]  = 010 (CTC Mode)
	// COM0A[1:0] = 00
	// COM0B[1:0] = 00
	// CS0[2:0]   = 011 // Clock Select = Clock/64
	// FOC0A      = 0	// Force Output Compare A
	// FOC0B      = 0	// Force Output Compare B

	TCCR0A = 0b00000010;
	TCCR0B = 0b00000011;		// Clock/64
	OCR0A  = F_CPU/64/1000;		// 1msƂɊ荞(1000Hz)
	TIMSK  = 0b0000001;			// rAv荞ݗL

	sei();	// Enable Interrupt
}

/******************************************************************************
	GPIO INITIALIZE
******************************************************************************/
void initGPIO()
{
	// Global Pull Up Enable for All Input Pins
	MCUCR &= ~0x10;

	// PORT DIR ( 0:Input / 1:Output)
	DDRB  = 0b11111111;
	DDRD  = 0b01110000;

	// PULLUP for INPUT (1:PullUp)
	PORTB = 0b11111111;
	PORTD = 0b11111111;
}

/******************************************************************************
	KEY
******************************************************************************/
#define KEY_START	0x01
#define KEY_STOP1	0x02
#define KEY_STOP2	0x04
#define KEY_STOP3	0x08

uint8_t readKey(void)
{
	static uint8_t	prevKey = 0xFF;
	uint8_t	key = ~(PIND & 0x0F);
	uint8_t	diff= prevKey ^ key;
	prevKey = key;
	return key & diff;
}

/******************************************************************************
	MAIN
******************************************************************************/
#define MODE_MANUALSTOP	0
#define MODE_AUTOSTOP	1
#define MODE_FREERUN	2

#define START_SPEED	50
#define STOP_SPEED	10
#define BREAK_SPEED	32

uint8_t	wait[3];
uint8_t	speed[3];
uint8_t	count[3];
uint8_t	slot[3];

uint8_t moveSlot(uint8_t mode)
{
	uint8_t	n;
	for(n=0; n<3; n++){
		count[n] += speed[n];
		if( count[n] < 150 ) continue;
		count[n] -= 150;
		slot[n] = (slot[n] + 1) % 50;

		switch(mode){
			case MODE_AUTOSTOP:
				if( speed[n] > 0 && wait[n] == 0 ) speed[n]--;
				if( speed[n] < STOP_SPEED && (slot[n]%5) == 2) speed[n] = 0;
				break;
			case MODE_MANUALSTOP:
				if( speed[n] > 0 && speed[n] <= BREAK_SPEED ) speed[n]--;
				if( speed[n] <= BREAK_SPEED-1 && (slot[n]%5) == 2) speed[n] = 0;
				break;
			case MODE_FREERUN:
				break;
		}
	}

	if( speed[0]==0 && speed[1]==0 && speed[2]==0 ){
		return 1;
	}else{
		return 0;
	}
}

void controlSlot(void)
{
	static uint8_t	state = 0;
	static uint8_t	blinkCounter = 0;
	static uint8_t	freerunWait = 0;

	uint8_t	key = readKey();
	uint8_t	blink = 0;

	switch(state){
		case 0:	// STOP
			blinkCounter++;
			if( (blinkCounter & 0x10) && slot[0]==slot[1] && slot[1]==slot[2] ){
				blink = 1;
			}
			if( key & KEY_START ) state = 1;
			break;
		case 1:	// START
			freerunWait = 40;
			speed[0] = START_SPEED + 12;
			speed[1] = START_SPEED +  6;
			speed[2] = START_SPEED +  0;
			wait[0] = 64;
			wait[1] = 32;
			wait[2] =  0;
			state = 2;
			break;
		case 2:	// ROTATING
			moveSlot( MODE_FREERUN );
			freerunWait--;
			if( freerunWait == 0 ) state = 3;
			break;
		case 3:	// ROTATING
			if( (key & KEY_STOP1) && speed[0] > BREAK_SPEED) speed[0] = BREAK_SPEED;
			if( (key & KEY_STOP2) && speed[1] > BREAK_SPEED) speed[1] = BREAK_SPEED;
			if( (key & KEY_STOP3) && speed[2] > BREAK_SPEED) speed[2] = BREAK_SPEED;
			if( moveSlot( MODE_MANUALSTOP ) ) state = 0;
			if( key & KEY_START) state = 4;
			break;
		case 4:	// AUTO STOP
			if( wait[0] > 0 ) wait[0]--;
			if( wait[1] > 0 ) wait[1]--;
			if( wait[2] > 0 ) wait[2]--;
			if( moveSlot( MODE_AUTOSTOP ) ) state = 0;
			break;
	}

	if( blink == 0 ){
		g_DispBuff[0] = DRUM[slot[0]];
		g_DispBuff[1] = DRUM[slot[1]];
		g_DispBuff[2] = DRUM[slot[2]];
	}
	else{
		g_DispBuff[0] = 0xFF;
		g_DispBuff[1] = 0xFF;
		g_DispBuff[2] = 0xFF;
	}
}

int main(void)
{
	uint8_t prevTimerCounter = 0;

	initGPIO();
	initTimer();

	slot[0] = slot[1] = slot[2] = 2;

	while(1){
		while( prevTimerCounter == timerCounter20ms );
		prevTimerCounter = timerCounter20ms;
		controlSlot();
	}
}
